
% Script for collecting inertial calibration data in combination with a
% calibration body providing a number of orientation for static placement
% of the IMU array. Follow the collection, the calibration parameters
% (scale factor,array alignment,internal alignment,bias) are estimated and
% a calibration matrix for the whole array is generated.
% Author: John-Olof Nilsson, Isaac Skog

% Clear workspace
clear all, close all, clc
format compact

% Parameters
internalMisalignment = true;    % Model internal misalignment or not
target_board = 'MIMU22BT';
nr_sides = 20;                  % Number of sides on calibration body
global g_local;
g_local = gravity(59,0); % Stockholm (latitude, altitude)
scale_acc  = 1/2048*9.80665;
scale_gyro = 1/16.4;
calib_gain = 2^15;

% Open serial port (Change adress to that provide by system)
com_port = 'COM36';
com = serial(com_port, 'BaudRate', 460800,'InputBufferSize',10000);
fopen(com);

% Prepare figure
figure(1),clf
set(gcf,'WindowStyle','docked')
subplot(2,1,1)
title('Accelerometer readings');
xlabel('sample number')
ylabel('a [m/s^2]');
grid on
subplot(2,1,2)
title('Gyroscope readings');
xlabel('sample number')
ylabel('\omega [deg/s]');
grid on

% Target board specific parameters
wireless = true;
if strcmp(target_board,'MIMU4444');
    wireless = false;
    nr_imus = 32;
    C_nom = [1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1];
    %command = [40 191 255 255 255 1 3 229];
    command = [40 255 255 255 255 1 4 37];
elseif strcmp(target_board,'MIMU3333')
    wireless = false;
    nr_imus = 14;
    C_nom = [1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1];
    command = [40 0 3 155 127 1 1 70];
elseif strcmp(target_board,'MIMU22BT')
    wireless = true;
    nr_imus = 4;
    C_nom = [1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1;
        1 0 0   0 1 0   0 0 1   1 0 0   0 1 0   0 0 1;
        0 -1 0  -1 0 0  0 0 -1  0 -1 0  -1 0 0  0 0 -1];
    command = [40 0 0 0 15 4 0 59];
else
    disp('Target board not recognized. Aborting calibration')
    fclose(com);
    exit;
end
data_length = 0; %[s]
if wireless
    data_length = 5;
else
    data_length = 2;
end

disp(' ');
disp('Script for logging calibration data sets with wireless MIMU platform and icosahedron rig.');
disp('Please verify that the platform is warmed up');
disp(' ');

input('Place body with side 1 up and press any key','s');

% Loop over the body sides
calibration_measurements = cell(nr_sides,1);
side_counter=1;
while true
    % This is to let potential vibrations from the key press die out
    pause(1);

    % Tell OpenShoe to output inertial data
    fwrite(com,command,'uint8');
    fread(com,4,'uint8');
        
    % Open binary file for saving inertial data
    file = fopen('imu_data.bin','w');
    
    % Logg data_length amount of data
    t_end=cputime+data_length;
    while t_end-cputime>0
        if com.BytesAvailable>0
            fwrite(file,fread(com,com.BytesAvailable,'uint8'),'uint8');
        end
    end
    % Stop output and close file
    fwrite(com,[33 0 33],'uint8');
    fclose(file);
    
    % Parse recorded data
    if wireless
        inertial_data=matlab_parse_wireless('imu_data.bin',uint8(nr_imus));
    else
        inertial_data=matlab_parse_bin('imu_data.bin',uint8(nr_imus));
    end
    % Add data to estimation data structure
    calibration_measurements{side_counter} = inertial_data;

    % Display data
    inertial_data=double(inertial_data);
    figure(1),clf
    handle=subplot(2,1,1);
    cla(handle), hold on
    title('Accelerometer readings');
    xlabel('sample number')
    ylabel('a [m/s^2]');
    for i=0:nr_imus-1
        R = reshape(C_nom(i+1,1:9),3,3);
        data = R*inertial_data(i*6+(1:3),:);
        plot(data'*scale_acc)
    end
    handle=subplot(2,1,2);
    cla(handle), hold on
    title('Gyroscope readings');
    xlabel('sample number')
    ylabel('\omega [deg/s]');
    for i=0:nr_imus-1
        R = reshape(C_nom(i+1,1:9),3,3);
        data = R*inertial_data(i*6+(4:6),:);
        plot(data'*scale_gyro)
    end
    
    % Done with side
    side_counter=side_counter+1;
    if side_counter<=nr_sides
        select = input(sprintf('Reject[r]/abort[a]/accept[any other key] (and place body with side %i up)',side_counter),'s');
        if strcmp(select,'r')
            side_counter=side_counter-1;
        end
        if strcmp(select,'a')
            disp('Calibration was aborted.')
            break;
        end
    else
        select = input('Reject[r]/finish[any other key]','s');
        if strcmp(select,'a')
            disp('Calibration was aborted.')
            break;
        end
        if strcmp(select,'r')
            side_counter=side_counter-1;
        else
            % Estimate and save the calibration parameters
            disp(' ');
            disp('Estimating calibration parameters (this may take a while)');
            [theta,y_cal_mean]=MIMU_calib_param_est(calibration_measurements,nr_imus,C_nom,scale_acc,scale_gyro,nr_sides,internalMisalignment);
            disp(' ');
            disp('Calibration data logging has finished');
            disp('Calibration result:');
            disp(theta);
            % Calculate and generate code for calibration matrix for 
            [C_int,b_acc_int,b_gyro_int] = calibration_matrix(theta,1/scale_acc,1/scale_gyro,C_nom,calib_gain);
            code_generation(C_int,b_acc_int,b_gyro_int,nr_imus);
            disp('Calibration matrix:');
            disp(b_acc_int);
            disp(b_gyro_int);
            disp(C_int);
            % Plot calibration data (with and without compensation for
            % calibration)
            disp('Plotting calibration data');
            MIMU_plot_calib(calibration_measurements,nr_imus,C_int,C_nom,theta,scale_acc,scale_gyro,calib_gain,b_acc_int,b_gyro_int,nr_sides);
            % Save workspace 
            filename = sprintf('./calibration_%02i%02i%02i_%02i%02i%02i',fix(clock()));
            save(filename);
            disp(['The data and the result is saved in ' filename]);
            break
        end
    end
    
    % 'Flush' buffers
    fread(com,com.BytesAvailable);    
end

fclose(com);

